"""This module contains the worker class calculating the visualizations."""

from __future__ import annotations

import math
import os
import shutil
import signal
import subprocess
from threading import Thread, Event
import time
from typing import Dict, Optional, NamedTuple, Tuple, TypedDict, cast

from django.db import connection

from django.conf import settings as conf
from core import redis
from core.tasks import app
from core.lights import controller, lights
from core.lights import leds
from core.lights.ring import Ring
from core.lights.wled import WLED
from core.lights.strip import Strip
from core.lights.screen import Screen
from core.lights.device import Device
from core.lights.programs import LedProgram, LightProgram, ScreenProgram
from core.lights.programs import Alarm, Cava, Disabled
from core.lights.led_programs import Adaptive, Fixed, Rainbow
from core.lights.exceptions import ScreenProgramStopped
from core.lights.screen_programs import Visualization, Video
from core.settings import storage
from core.settings.storage import DeviceBrightness, DeviceMonochrome, DeviceProgram
from core.util import optional

lights_lock = redis.connection.lock("lights_lock")


class Settings(TypedDict):
    """A type containing all settings affecting multiple programs."""

    ups: float
    dynamic_resolution: bool
    program_speed: float
    fixed_color: Tuple[float, float, float]
    last_fixed_color: Tuple[float, float, float]


class Utilities(NamedTuple):
    """A type containing all utility programs that can not run directly on a device."""

    disabled: Disabled
    cava: Cava
    alarm: Alarm


class Devices(NamedTuple):
    """A type containing all available devices."""

    ring: Ring
    strip: Strip
    wled: WLED
    screen: Screen


def start() -> None:
    """Initializes this module by starting the lights loop."""
    _loop.delay()
    connection.close()


class DeviceManager:
    """A class managing all visualization devices.
    This class maintains state, but only in the worker thread that updates the lights.
    This keeps the necessary variables local to the thread, avoiding db/redis queries each frame."""

    def __init__(self) -> None:

        self.loop_active: Optional[Event] = Event()

        self.devices = Devices(Ring(self), Strip(self), WLED(self), Screen(self))

        # these settings are mirrored from the database,
        # because some of them are accessed multiple times per update.
        self.settings: Settings = {
            "ups": storage.get("ups"),
            "dynamic_resolution": storage.get("dynamic_resolution"),
            "program_speed": storage.get("program_speed"),
            "fixed_color": storage.get("fixed_color"),
            "last_fixed_color": storage.get("fixed_color"),
        }

        self.utilities = Utilities(Disabled(self), Cava(self), Alarm(self))
        cava_installed = shutil.which("cava") is not None

        # a dictionary containing all led programs by their name
        led_programs: Dict[str, LedProgram] = {
            self.utilities.disabled.name: self.utilities.disabled
        }
        led_program_classes = [Fixed, Rainbow]
        if cava_installed:
            led_program_classes.append(Adaptive)
        for led_program_class in led_program_classes:
            led_program = led_program_class(self)
            led_programs[led_program.name] = led_program
        # a dictionary containing all screen programs by their name
        screen_programs: Dict[str, ScreenProgram] = {
            self.utilities.disabled.name: self.utilities.disabled
        }
        logo_loop = Video(self, "LogoLoop.mp4", loop=True)
        screen_programs[logo_loop.name] = logo_loop
        if cava_installed:
            for variant in sorted(Visualization.get_variants()):
                screen_programs[variant] = Visualization(self, variant)

        redis.put("led_programs", list(led_programs.keys()))
        redis.put("screen_programs", list(screen_programs.keys()))

        # a dictionary containing *all* programs by their name
        self.programs: Dict[str, LightProgram] = {**led_programs, **screen_programs}

        for device in self.devices:
            device.load_program()

        self.consumers_changed()

        self.listener = Thread(target=self.listen_for_changes)
        self.listener.start()

    def listen_for_changes(self) -> None:
        """Listens for changed settings on the redis channel and acts accordingly."""
        pubsub = redis.connection.pubsub(ignore_subscribe_messages=True)
        pubsub.subscribe("lights_settings_changed")
        for message in pubsub.listen():
            settings_changed = message["data"]

            if settings_changed == "stop":
                # delete the lock the main thread is checking each loop in order to stop it
                # this removes the need for an extra redis variable
                # that would need to be checked every loop
                if self.loop_active is not None:
                    self.loop_active.set()
                    self.loop_active = None
                break

            if settings_changed == "alarm_started":
                self.alarm_started()
                continue
            if settings_changed == "alarm_stopped":
                self.alarm_stopped()
                continue

            # flush the cache before accessing the database so no stale data is read
            storage.cache.clear()

            if settings_changed == "adjust_screen":
                self.devices.screen.adjust()
                # restart the screen program after a resolution change so it fits the screen again
                if storage.get("initial_resolution") != self.devices.screen.resolution:
                    # changing resolutions takes a while, wait until it was applied
                    self.restart_screen_program(sleep_time=2)
                else:
                    self.restart_screen_program()
                continue

            if settings_changed == "base":
                # base settings affecting every device were modified
                if not math.isclose(self.settings["ups"], storage.get("ups")):
                    # only update ups if they actually changed,
                    # because cava and the screen program need to be restarted
                    old_ups = self.settings["ups"]
                    self.settings["ups"] = storage.get("ups")
                    self.set_cava_framerate()
                    self.restart_screen_program(sleep_time=1 / old_ups * 5)
                self.settings["dynamic_resolution"] = storage.get("dynamic_resolution")
                self.settings["fixed_color"] = storage.get("fixed_color")
                self.settings["program_speed"] = storage.get("program_speed")
                continue

            # a device was changed
            # reload all settings for this device from the database
            # instead of communicating the exact changed setting.
            device_name = settings_changed
            device = getattr(self.devices, device_name)
            assert device_name in ["ring", "strip", "wled", "screen"]
            device.brightness = storage.get(
                cast(DeviceBrightness, f"{device_name}_brightness")
            )
            device.monochrome = storage.get(
                cast(DeviceMonochrome, f"{device_name}_monochrome")
            )
            if device_name == "wled":
                self.devices.wled.ip = storage.get("wled_ip")
                self.devices.wled.port = storage.get("wled_port")
            program = self.programs[
                storage.get(cast(DeviceProgram, f"{device_name}_program"))
            ]
            self.set_program(device, program)
        connection.close()

    def set_program(
        self, device: Device, program: LightProgram, has_lock=False
    ) -> None:
        """Changes the program of the given device to the given program."""
        # don't allow program change on disconnected devices
        if not device.initialized:
            return

        with optional(not has_lock, lights_lock):
            if device.program == program:
                # nothing to do
                return

            device.program.release()
            # see explanation in except ScreenProgramStopped why we don't always use() here
            if not (
                isinstance(
                    self.programs[storage.get("last_screen_program")], Visualization
                )
                and isinstance(
                    self.programs[storage.get("screen_program")], Visualization
                )
            ):
                program.use()

            device.program = program
            self.consumers_changed()

            if program.name == "Disabled":
                device.clear()

        # Disable the pwr led if the ring is active.
        # The pwr led ruins the clean look of a ring spectrum,
        # and an active led ring is enough of an indicator that the Pi is running.
        if device.name == "ring":
            if program.name == "Disabled":
                leds.enable_pwr_led()
            else:
                leds.disable_pwr_led()

    def consumers_changed(self) -> None:
        """Stops the loop if no led is active, starts it otherwise"""
        if self.utilities.disabled.consumers == 4:
            assert self.loop_active
            self.loop_active.clear()
            redis.put("lights_active", False)
        else:
            assert self.loop_active
            self.loop_active.set()
            redis.put("lights_active", True)

    def alarm_started(self) -> None:
        """Makes alarm the current program but doesn't update the database."""
        self.utilities.alarm.use()

        leds.disable_pwr_led()

        self.settings["last_fixed_color"] = self.settings["fixed_color"]
        for device in [self.devices.ring, self.devices.wled, self.devices.strip]:
            self.set_program(device, self.programs["Fixed"])
        # the screen program adapts with the alarm and is not changed

    def alarm_stopped(self) -> None:
        """Restores the state from before the alarm."""
        self.utilities.alarm.release()
        self.settings["fixed_color"] = self.settings["last_fixed_color"]

        leds.enable_pwr_led()

        # the database still contains the program from before the alarm. restore it.
        for device in [self.devices.ring, self.devices.wled, self.devices.strip]:
            self.set_program(
                device,
                self.programs[
                    storage.get(cast(DeviceProgram, f"{device.name}_program"))
                ],
            )

    def restart_screen_program(self, sleep_time=None, has_lock=False) -> None:
        """Restarts the current screen program. Waits sleep_time seconds in the Disabled state."""
        if self.devices.screen.program == self.utilities.disabled:
            return
        screen_program = self.devices.screen.program
        self.set_program(
            self.devices.screen, self.utilities.disabled, has_lock=has_lock
        )
        # wait until the program was computed, so it executed its shutdown
        if sleep_time:
            time.sleep(sleep_time)
        else:
            time.sleep(1 / self.settings["ups"] * 5)
        self.set_program(self.devices.screen, screen_program, has_lock=has_lock)

    def set_cava_framerate(self) -> None:
        """Update the cava.config file to contain the current framerate."""
        subprocess.call(
            [
                "sed",
                "-i",
                "-r",
                "-e",
                f"s/(^framerate\\s*=).*/\\1 {self.settings['ups']}/",
                os.path.join(conf.BASE_DIR, "config/cava.config"),
            ]
        )
        if self.utilities.cava.cava_process:
            self.utilities.cava.cava_process.send_signal(signal.SIGUSR1)

    def _set_led_colors(self) -> None:
        assert isinstance(self.devices.ring.program, LedProgram)
        assert isinstance(self.devices.strip.program, LedProgram)
        assert isinstance(self.devices.wled.program, LedProgram)
        if self.devices.ring.program.name != "Disabled":
            if self.devices.ring.monochrome:
                ring_colors = [
                    self.devices.ring.program.strip_color()
                    for _ in range(self.devices.ring.LED_COUNT)
                ]
            else:
                ring_colors = self.devices.ring.program.ring_colors()
            self.devices.ring.set_colors(ring_colors)

        if self.devices.strip.program.name != "Disabled":
            strip_color = self.devices.strip.program.strip_color()
            self.devices.strip.set_color(strip_color)

        if self.devices.wled.program.name != "Disabled":
            if self.devices.wled.monochrome:
                wled_colors = [
                    self.devices.wled.program.strip_color()
                    for _ in range(self.devices.wled.led_count)
                ]
            else:
                wled_colors = self.devices.wled.program.wled_colors()
            self.devices.wled.set_colors(wled_colors)

    def loop(self) -> None:
        """The main lights loop. When active, compute every active program and set the devices."""
        while True:
            try:
                self.loop_active.wait()  # type: ignore[union-attr]
            except AttributeError:
                # the lock was deleted by the listener thread in order to stop the main thread
                # The display must be stopped from this thread, otherwise no new one can be created
                self.devices.screen.program.stop()
                self.listener.join()
                break

            computation_start = time.time()

            with lights_lock:
                # these programs only actually do work if their respective programs are active
                self.utilities.cava.compute()
                self.utilities.alarm.compute()

                self.devices.ring.program.compute()
                if self.devices.wled.program != self.devices.ring.program:
                    self.devices.wled.program.compute()
                if self.devices.strip.program != self.devices.ring.program:
                    self.devices.strip.program.compute()

                self._set_led_colors()

                try:
                    self.devices.screen.program.compute()
                except ScreenProgramStopped:
                    # If the program changes from one Visualization program to another,
                    # the new program will set active to true immediately,
                    # but the old program will set active to false only during the next frame.
                    # This would mean that active will stay false even though rendering takes place.
                    # We prevent this by starting the new program only after the old one stopped.
                    # This happens when this exception was catched
                    # and we switched from one Visualization program to another.
                    #
                    # This exception is also thrown if the window is closed.
                    # For these cases, set the program to disabled.
                    # This means that if the windows *is* closed
                    # after switching from one Visualization program to another,
                    # the program will not be set to disabled. This is a corner case we accept.
                    # It would be fixed by introducing a variable
                    # and checking whether such a change occurred in the recent past.
                    if isinstance(
                        self.programs[storage.get("last_screen_program")], Visualization
                    ) and isinstance(
                        self.programs[storage.get("screen_program")], Visualization
                    ):
                        self.programs[storage.get("screen_program")].use()
                    else:
                        self.set_program(
                            self.devices.screen, self.utilities.disabled, has_lock=True
                        )
                        controller.persist_program_change("screen", "Disabled")
                        lights.update_state()

            computation_time = time.time() - computation_start
            try:
                time.sleep(1 / self.settings["ups"] - computation_time)
            except ValueError:
                pass


@app.task
def _loop() -> None:
    manager = DeviceManager()
    manager.loop()
    connection.close()



"""This module registers the core apps websocket patterns."""

from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from django.core.asgi import get_asgi_application

import core.routing

APPLICATION = ProtocolTypeRouter(
    {
        "websocket": AuthMiddlewareStack(URLRouter(core.routing.WEBSOCKET_URLPATTERNS)),
        # using django's asgi application for http requests allows async processing
        # since django 4.0 due to context aware sync_to_async
        # https://github.com/django/django/pull/13882
        "http": get_asgi_application(),
    }
)



"""This module provides app wide utility functions."""
import subprocess
from contextlib import contextmanager
from typing import List, Tuple, ContextManager, Generator

from django.http import (
    HttpResponseForbidden,
    HttpResponseBadRequest,
    HttpResponse,
    QueryDict,
)


def strtobool(value: str) -> bool:
    """Convert a string representation of a boolean to True or False."""
    value = value.lower()
    if value in ("y", "yes", "t", "true", "on", "1"):
        return True
    if value in ("n", "no", "f", "false", "off", "0"):
        return False
    raise ValueError(f"invalid truth value {value}")


@contextmanager
def optional(condition: bool, context_manager: ContextManager) -> Generator:
    """Conditionally apply a context_manager. If condition is false, no context manager is used."""
    if condition:
        with context_manager:
            yield
    else:
        yield


def camelize(snake_dict: dict) -> dict:
    """Transforms each key of the given dict from snake_case to CamelCase."""

    def camelize_str(snake: str) -> str:
        """Transforms a CamelCase string into a snake_case string."""
        head, *tail = snake.split("_")
        return head + "".join(word.capitalize() for word in tail)

    return {camelize_str(k): v for k, v in snake_dict.items()}


def extract_value(request: QueryDict) -> Tuple[str, HttpResponse]:
    """Return the "value" from the given QueryDict,
    and an HttpResponse if value is not None, or HttpResponseBadRequest if it is None."""
    value = request.get("value")
    if value is None:
        return "", HttpResponseBadRequest("value must be specified")
    return value, HttpResponse()


def get_devices() -> List[str]:
    """Returns a list of all network devices."""
    output = subprocess.check_output(
        "ip route show default".split(), universal_newlines=True
    )
    words = output.split()
    devices = []
    for cur, nex in zip(words, words[1:]):
        if cur == "dev":
            devices.append(nex)
    if not devices:
        raise ValueError("no devices found")
    return devices


def ip_of_device(device: str) -> str:
    """Returns the IP that the system has on the given network device."""
    output = subprocess.check_output(
        f"ip -4 a show dev {device}".split(), universal_newlines=True
    )
    ip = None
    for line in output.split("\n"):
        line = line.strip()
        if not line.startswith("inet"):
            continue
        ip = line.split()[1].split("/")[0]
        break
    if not ip:
        raise ValueError(f"ip not found for {device}")
    return ip


def broadcast_of_device(device: str) -> str:
    """Returns the broadcast address of the given network device."""
    output = subprocess.check_output(
        f"ip -o -f inet addr show {device}".split(), universal_newlines=True
    )
    words = output.split()
    return words[5]


def service_installed(service: str) -> bool:
    """Returns whether the given systemd service is installed."""
    if not service.endswith(".service"):
        service += ".service"
    try:
        out = subprocess.check_output(
            ["systemctl", "list-unit-files", service], text=True
        )
    except subprocess.CalledProcessError:
        return False
    return len(out.splitlines()) > 3


def csrf_failure(  # pylint: disable=unused-argument
    request, reason="", template_name=""
) -> HttpResponseForbidden:
    """A custom csrf failure view that fits inside a toast message."""
    return HttpResponseForbidden("Please reload")


def format_resolution(resolution: Tuple[int, int]) -> str:
    """Format an int-tuple as a readable resolution."""
    return f"{resolution[0]}x{resolution[1]}"



"""This module handles playback flow of music."""

from __future__ import annotations

import datetime
import logging
import os
import random
import time
from typing import Optional, Tuple

from django.conf import settings as conf
from django.db import connection, transaction
from django.utils import timezone

from core import models, redis, user_manager
from core.lights import controller as lights_controller
from core.musiq import musiq
from core.settings import storage
from core.tasks import app
from core.musiq import song_utils
from core.models import CurrentSong

queue_changed = redis.Event("queue_changed")
buzzer_stopped = redis.Event("buzzer_stopped")

queue = models.QueuedSong.objects


class PlaybackError(Exception):
    pass


def start() -> None:
    """Initializes this module by starting the playback and buzzer loop."""
    paused = storage.get("paused")
    redis.put("paused", paused)
    _handle_buzzer.delay()
    _loop.delay()


class Playback:
    """Class containing all playback related methods."""

    def __init__(self):
        from core.musiq.fake_player import FakePlayer

        redis.put("playing", False)

        queue.delete_placeholders()

        self.players = {
            "fake": FakePlayer(),
        }

        if redis.get("spotify_available"):
            from core.musiq.spotify_player import SpotifyPlayer

            self.players["spotify"] = SpotifyPlayer()
        if redis.get("mopidy_available"):
            from core.musiq.mopidy_player import MopidyPlayer

            self.players["mopidy"] = MopidyPlayer()

        if redis.get("spotify_available") and storage.get("output").startswith(
            "spotify-"
        ):
            redis.put("active_player", "spotify")
        elif not redis.get("mopidy_available"):
            redis.put("active_player", "fake")
        else:
            redis.put("active_player", "mopidy")

    def player(self):
        return self.players[redis.get("active_player")]

    def play_alarm(self, interrupt=False, from_buzzer=True) -> None:
        """Play the alarm sound. If specified, interrupts the currently playing song."""
        redis.put("alarm_playing", True)
        lights_controller.alarm_started()

        success_probability = storage.get("buzzer_success_probability")
        if success_probability >= 0 and from_buzzer:
            if random.random() <= success_probability:
                folder = "resources/sounds/yes"
            else:
                folder = "resources/sounds/no"
            choice = random.choice(os.listdir(os.path.join(conf.BASE_DIR, folder)))
            path = os.path.join(conf.BASE_DIR, folder, choice)
        else:
            path = os.path.join(conf.BASE_DIR, "resources/sounds/alarm.m4a")
        duration = song_utils.get_metadata(path)["duration"]

        redis.put("alarm_duration", duration)

        self.player().play_alarm(interrupt, path)

        musiq.update_state()

        # TODO: is it a very bad idea to simply sleep here for the duration of the alarm?
        # self._wait_until_song_end()
        time.sleep(duration)

        lights_controller.alarm_stopped()
        redis.put("alarm_playing", False)

        if not interrupt:
            # if no song immediately continues playing, a manual state update is needed
            musiq.update_state()

    def _get_next_song(self) -> Tuple[Optional[models.CurrentSong], bool]:
        """Returns the next song that should be played, or None if no song should be played.
        Additionally returns whether the song was recovered from the database (True)
        or dequeued from the queue (False)."""
        # TODO: is this playback_started clear necessary?
        # self.playback_started.clear()

        if models.CurrentSong.objects.exists():
            # recover interrupted song from database
            return models.CurrentSong.objects.get(), True

        if queue.count() == 0:
            queue_changed.wait()
            queue_changed.clear()

            # restart the loop to check again whether songs are available
            # in case of a false wakeup this causes as to wait again
            return None, False

        # select the next song depending on settings
        song: Optional[models.QueuedSong]
        if storage.get("interactivity") in [
            storage.Interactivity.upvotes_only,
            storage.Interactivity.full_voting,
        ]:
            with transaction.atomic():
                song = queue.confirmed().order_by("-votes", "index")[0]
                song_id = song.id
                queue.remove(song.id)
        elif storage.get("shuffle"):
            confirmed = queue.confirmed()
            index = random.randint(0, confirmed.count() - 1)
            song_id = confirmed[index].id
            song = queue.remove(song_id)
        else:
            # move the first song in the queue into the current song
            song_id, song = queue.dequeue()

        if song is None:
            # either the semaphore didn't match up with the actual count
            # of songs in the queue or a race condition occured
            logging.warning("dequeued on empty list")
            return None, False

        if song.internal_url == "alarm":
            self.play_alarm()
            return None, False

        # stop backup stream.
        # when the dequeued song starts playing, the backup stream playback is stopped
        redis.put("backup_playing", False)

        assert song.internal_url
        current_song = models.CurrentSong.objects.create(
            queue_key=song_id,
            manually_requested=song.manually_requested,
            votes=song.votes,
            internal_url=song.internal_url,
            external_url=song.external_url,
            stream_url=song.stream_url,
            artist=song.artist,
            title=song.title,
            duration=song.duration,
        )

        handle_autoplay()

        try:
            archived_song = models.ArchivedSong.objects.get(
                url=current_song.external_url
            )
            votes: Optional[int]
            if storage.get("interactivity") in [
                storage.Interactivity.upvotes_only,
                storage.Interactivity.full_voting,
            ]:
                votes = current_song.votes
            else:
                votes = None
            if storage.get("logging_enabled"):
                models.PlayLog.objects.create(
                    song=archived_song,
                    manually_requested=current_song.manually_requested,
                    votes=votes,
                )
        except (
            models.ArchivedSong.DoesNotExist,
            models.ArchivedSong.MultipleObjectsReturned,
        ):
            pass

        return current_song, False

    @staticmethod
    def _catch_up(current_song: models.CurrentSong, recovered: bool) -> Optional[int]:
        catch_up = None

        if recovered:
            # continue with the current song (approximately) where we last left
            if redis.get("paused"):
                catch_up = round(
                    (current_song.last_paused - current_song.created).total_seconds()
                    * 1000
                )
            else:
                catch_up = round(
                    (timezone.now() - current_song.created).total_seconds() * 1000
                )
            if catch_up > current_song.duration * 1000:
                catch_up = -1

        return catch_up

    def _wait_until_song_end(self) -> bool:
        """Wait until the song is over.
        Returns True when finished without errors, False otherwise."""
        # This is the event based approach. Unfortunately too error-prone.
        # If mopidy crashes/restarts for example, no track_playback_ended event is sent
        # playback_ended.wait()
        error = False
        while True:
            current_song = CurrentSong.objects.get()
            try:
                if self.player().should_stop_waiting(error):
                    break
            except PlaybackError:
                error = True
            paused = redis.get("paused")
            if paused:
                # stay in the loop, the song won't end while paused
                # progress = (current_song.last_paused - current_song.created).total_seconds()
                pass
            else:
                progress = (timezone.now() - current_song.created).total_seconds()
                if progress >= current_song.duration:
                    break
            # use internal timekeeping to decide when a song is over to lower mopidy performance use
            # current_song = models.CurrentSong.objects.get()
            # if (timezone.now() - current_song.created).total_seconds() > current_song.duration:
            #    break
            time.sleep(0.1)
            if redis.get("stop_playback_loop"):
                # in order to stop the playback thread, return False, making the main loop restart.
                # it will check this variable again and terminate itself.
                return False
            if redis.get("alarm_requested"):
                redis.put("alarm_requested", False)
                self.play_alarm(interrupt=True)
                # the current song was interrupted and needs to be resumed at the correct position
                # returning False will notify the main loop about this interruption,
                # making it restart the song correctly
                # we don't want the song to skip over the time when the alarm was playing
                # thus, we offset the creation date of the current song by the length of the alarm
                # Warning: if this duration does not fit the duration of the actual alarm,
                # Raveberry's internal state gets desynced and weird errors happen
                current_song.created += datetime.timedelta(
                    seconds=redis.get("alarm_duration")
                )
                current_song.save()

                return False
        return not error

    def _song_finished(self, current_song: models.CurrentSong) -> None:
        """Handles things that might need to happen after a song ended,
        e.g.repeat, alarm and backup stream."""
        if storage.get("repeat"):
            queue.enqueue(
                {
                    "artist": current_song.artist,
                    "title": current_song.title,
                    "duration": current_song.duration,
                    "internal_url": current_song.internal_url,
                    "external_url": current_song.external_url,
                    "stream_url": current_song.stream_url,
                },
                False,
                enqueue_first=False,
            )
            queue_changed.set()

        if user_manager.partymode_enabled() and random.random() < storage.get(
            "alarm_probability"
        ):
            self.play_alarm(from_buzzer=False)

        if not queue.exists() and storage.get("backup_stream"):
            redis.put("backup_playing", True)
            self.player().play_backup_stream()

        musiq.update_state()

    def loop(self) -> None:
        """The main loop of the player.
        Takes a song from the queue and plays it until it is finished."""

        while True:
            if redis.get("playback_error"):
                # sleep for a short while so continuing errors don't lead to busy loops
                time.sleep(0.5)
                continue

            if redis.get("stop_playback_loop"):
                break

            current_song, recovered = self._get_next_song()
            if current_song is None:
                continue

            catch_up = self._catch_up(current_song, recovered)

            try:
                self.player().start_song(current_song, catch_up)
            except PlaybackError:
                # when a song can't be started, pause the playback
                # and have the user restart playback manually after fixing the error
                musiq.controller._pause()
                musiq.update_state()
                continue
            redis.put("playing", True)

            musiq.update_state()

            # don't wait for the song to end if catch_up is negative (=the song should be skipped)
            if catch_up is None or catch_up >= 0:
                if not self._wait_until_song_end():
                    # there was an error while waiting for the song to end
                    # This happens when we could not connect to mopidy (ConnectionError)
                    # or when an interrupting alarm was initiated.
                    # we do not delete the current song but recover its state by restarting the loop
                    storage.put("paused", False)
                    redis.put("paused", False)
                    redis.put("playing", False)
                    continue
            # Allowing new songs to start playing while paused introduces many edge cases
            # Instead of dealing with them, always start playback when skipping a paused song
            storage.put("paused", False)
            redis.put("paused", False)
            redis.put("playing", False)

            current_song.delete()

            self._song_finished(current_song)


@app.task
def _loop() -> None:
    playback = Playback()
    playback.loop()
    connection.close()


@app.task
def _handle_buzzer() -> None:
    try:
        import gpiozero
    except ModuleNotFoundError:
        return
    buzzer = gpiozero.Button(16)

    def on_press():
        trigger_alarm()

    buzzer.when_pressed = on_press

    # wait until this task is told to exit
    buzzer_stopped.wait()


def trigger_alarm() -> bool:
    """Initiate an alarm."""
    # do not allow the alarm to be triggered too frequently
    now = time.time()
    if (now - redis.get("last_buzzer")) < storage.get("buzzer_cooldown"):
        logging.warning("alarm triggered too quickly")
        return False
    # do not allow an alarm to be triggered while one is already playing
    # or when an alarm is currently in the process of being played
    if redis.get("alarm_playing") or redis.get("alarm_requested"):
        logging.warning("last alarm not yet finished")
        return False
    redis.put("last_buzzer", now)
    if redis.get("playing"):
        # if a song is currently playing, inform the loop waiting for the song to end
        # about this alarm. It will interrupt the current song and play the alarm
        redis.put("alarm_requested", True)
    else:
        # insert a special queue song to wake up the main loop and make it play the alarm
        queue.enqueue(musiq.get_alarm_metadata(), True)
        queue_changed.set()
    return True


def handle_autoplay(url: Optional[str] = None) -> None:
    """Checks whether to add a song by autoplay and does so if necessary.
    :param url: if given, this url is used to find the next autoplayed song.
    Otherwise, the current song is used."""
    if storage.get("autoplay") and models.QueuedSong.objects.count() == 0:
        if url is None:
            # if no url was specified, use the one of the current song
            try:
                current_song = models.CurrentSong.objects.get()
                url = current_song.external_url
            except (
                models.CurrentSong.DoesNotExist,
                models.CurrentSong.MultipleObjectsReturned,
            ):
                return

        from core.musiq.song_provider import SongProvider

        provider = SongProvider.create(external_url=url)
        try:
            suggestion = provider.get_suggestion()
            # The player loop is not restarted after error automatically.
            # As this function can raise several exceptions (it might do networking)
            # we catch every exception to make sure the loop keeps running
        except Exception as error:  # pylint: disable=broad-except
            logging.exception("error during suggestions for %s: %s", url, error)
        else:
            provider = SongProvider.create(external_url=suggestion)
            provider.request("", archive=False, manually_requested=False)


def stop() -> None:
    """Stops the playback main loop, only used for tests."""
    redis.put("stop_playback_loop", True)
    queue_changed.set()



# Generated by Django 4.1.7 on 2023-05-05 19:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0017_archivedquery_core_archivedquery_query_trgm_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="setting",
            name="value",
            field=models.CharField(max_length=1000),
        ),
    ]



"""This module interfaces with spotify."""

import logging
from contextlib import contextmanager
from typing import Iterator

import requests.exceptions
from spotipy import SpotifyException, SpotifyOauthError

from core.musiq.playback import PlaybackError
from core.musiq import player
from core.musiq.spotify import Spotify
from core.musiq import song_utils


@contextmanager
def spotify_api(reraise: bool = False) -> Iterator:
    """A context that should be used around every spotify api playback call.
    Catches common exceptions and deals with them."""
    try:
        yield
        # after a successful api call, there is no error with spotify
        player.set_playback_error(False)
    except (
        SpotifyException,
        SpotifyOauthError,
        requests.exceptions.ConnectionError,
    ) as e:
        logging.warning("Spotify API Error: %s", e)
        player.set_playback_error(True)
        if reraise:
            raise e


class SpotifyPlayer(player.Player):
    """Class containing methods to interface with Spotify."""

    def start_song(self, song, catch_up: float):
        if song_utils.determine_url_type(song.external_url) != "spotify":
            logging.warning("tried to play non-spotify song with spotify player")
            raise PlaybackError("Not a Spotify song")
        try:
            with spotify_api(reraise=True):
                # if catch_up is not None and catch_up >= 0:
                #    volume = Spotify.api.current_playback()["device"]["volume_percent"]
                #    Spotify.api.volume(0)
                Spotify.api.start_playback(uris=[song.internal_url])
                if catch_up is not None and catch_up >= 0:
                    Spotify.api.seek_track(catch_up)
                    # Spotify.api.volume(volume)
        except (
            SpotifyException,
            SpotifyOauthError,
            requests.exceptions.ConnectionError,
        ):
            raise PlaybackError("Spotify API error")

    def play_alarm(self, interrupt: bool, alarm_path: str) -> None:
        # since the sounds can not be played on Spotify, we only pause here to have
        # some time so they can be played some other way
        pause()

    def play_backup_stream(self):
        # Spotify can't play arbitrary internet streams
        pass


def restart() -> None:
    with spotify_api():
        Spotify.api.seek_track(0)


def seek_backward(seek_distance: float) -> None:
    with spotify_api():
        # TODO: Spotify often returns None in progress_ms. then we cannot seek
        current_position = Spotify.api.current_playback()["progress_ms"]
        if current_position:
            Spotify.api.seek_track(current_position - seek_distance * 1000)


def play() -> None:
    with spotify_api():
        Spotify.api.start_playback()


def pause() -> None:
    with spotify_api():
        Spotify.api.pause_playback()


def seek_forward(seek_distance: float) -> None:
    with spotify_api():
        current_position = Spotify.api.current_playback()["progress_ms"]
        if current_position:
            Spotify.api.seek_track(current_position + seek_distance * 1000)


def skip() -> None:
    with spotify_api():
        Spotify.api.next_track()


def set_volume(volume) -> None:
    with spotify_api():
        logging.error("trying to set volume")
        Spotify.api.volume(round(volume * 100))
        logging.error("finished to set volume")



"""This module handles all controls that change the playback."""

from __future__ import annotations

import datetime
import subprocess
from functools import wraps
from typing import Callable

from django.conf import settings as conf
from django.core.handlers.wsgi import WSGIRequest
from django.db import transaction
from django.db.models import F
from django.http import HttpResponseForbidden
from django.http.response import HttpResponse, HttpResponseBadRequest
from django.utils import timezone
from django.views.decorators.csrf import csrf_exempt

from core import models, redis, user_manager
from core.musiq import musiq, playback, player
from core.settings import storage
from core.util import extract_value

SEEK_DISTANCE = 10


def control(func: Callable) -> Callable:
    """A decorator for functions that control the playback.
    Every control changes the views state and returns an empty response.
    At least mod privilege is required during voting."""

    def _decorator(request: WSGIRequest) -> HttpResponse:
        if storage.get(
            "interactivity"
        ) != storage.Interactivity.full_control and not user_manager.has_controls(
            request.user
        ):
            return HttpResponseForbidden()
        response = func(request)
        musiq.update_state()
        if response is not None:
            return response
        return HttpResponse()

    return wraps(func)(_decorator)


def start() -> None:
    """Initializes this module by restoring the volume."""
    volume = storage.get("volume")
    _set_volume(volume)


@control
def restart(_request: WSGIRequest) -> None:
    """Restarts the current song from the beginning."""
    player.restart()
    try:
        current_song = models.CurrentSong.objects.get()
        current_song.created = timezone.now()
        current_song.save()
    except models.CurrentSong.DoesNotExist:
        pass


@control
def seek_backward(_request: WSGIRequest) -> None:
    """Jumps back in the current song."""
    player.seek_backward(SEEK_DISTANCE)
    try:
        current_song = models.CurrentSong.objects.get()
        now = timezone.now()
        current_song.created += datetime.timedelta(seconds=SEEK_DISTANCE)
        current_song.created = min(current_song.created, now)
        current_song.save()
    except models.CurrentSong.DoesNotExist:
        pass


@control
def play(_request: WSGIRequest) -> None:
    """Resumes the current song if it is paused.
    No-op if already playing."""
    player.play()
    try:
        # move the creation timestamp into the future for the duration of the pause
        # this ensures that the progress calculation (starting from created) is correct
        current_song = models.CurrentSong.objects.get()
        now = timezone.now()
        pause_duration = (now - current_song.last_paused).total_seconds()
        current_song.created += datetime.timedelta(seconds=pause_duration)
        current_song.created = min(current_song.created, now)
        current_song.save()
    except models.CurrentSong.DoesNotExist:
        pass
    storage.put("paused", False)
    redis.put("paused", False)


def _pause() -> None:
    player.pause()
    try:
        current_song = models.CurrentSong.objects.get()
        current_song.last_paused = timezone.now()
        current_song.save()
    except models.CurrentSong.DoesNotExist:
        pass
    storage.put("paused", True)
    redis.put("paused", True)


@control
def pause(_request: WSGIRequest) -> None:
    """Pauses the current song if it is playing.
    No-op if already paused."""
    _pause()


@control
def seek_forward(_request: WSGIRequest) -> None:
    """Jumps forward in the current song."""
    player.seek_forward(SEEK_DISTANCE)
    try:
        current_song = models.CurrentSong.objects.get()
        current_song.created -= datetime.timedelta(seconds=SEEK_DISTANCE)
        current_song.save()
    except models.CurrentSong.DoesNotExist:
        pass


def _skip() -> None:
    player.skip()
    redis.put("backup_playing", False)
    try:
        current_song = models.CurrentSong.objects.get()
        current_song.created = timezone.now() - datetime.timedelta(
            seconds=current_song.duration
        )
        current_song.save()
    except models.CurrentSong.DoesNotExist:
        pass


@control
def skip(_request: WSGIRequest) -> None:
    """Skips the current song and continues with the next one."""
    _skip()


@control
def set_shuffle(request: WSGIRequest) -> None:
    """Enables or disables shuffle depending on the given value.
    If enabled, a random song in the queue is chosen as the next one.
    If not, the first one is chosen."""
    enabled = request.POST.get("value") == "true"
    storage.put("shuffle", enabled)


@control
def set_repeat(request: WSGIRequest) -> None:
    """Enables or disables repeat depending on the given value.
    If enabled, a song is enqueued again after it finished playing."""
    enabled = request.POST.get("value") == "true"
    storage.put("repeat", enabled)


@control
def set_autoplay(request: WSGIRequest) -> None:
    """Enables or disables autoplay depending on the given value.
    If enabled and the current song is the last one,
    a new song is enqueued, based on the current one."""
    enabled = request.POST.get("value") == "true"
    storage.put("autoplay", enabled)
    playback.handle_autoplay()


def _set_volume(volume) -> None:
    try:
        # Try to set the volume via the pulse server.
        # This is faster and does not impact visualization
        subprocess.run(
            f"pactl set-sink-volume @DEFAULT_SINK@ {round(volume*100)}%".split(),
            env={"PULSE_SERVER": conf.PULSE_SERVER},
            check=True,
        )
    except (FileNotFoundError, subprocess.CalledProcessError):
        # pulse is not installed or there is no server running.
        # TODO: why does this hang with spotipy?
        # it can't change the volume on the phone, but it should simply raise an error which gets catched and then move on
        player.set_volume(volume)


@control
def set_volume(request: WSGIRequest) -> HttpResponse:
    """Sets the playback volume.
    value has to be a float between 0 and 1."""
    volume, response = extract_value(request.POST)
    _set_volume(float(volume))
    storage.put("volume", float(volume))
    return response


@control
def shuffle_all(request: WSGIRequest) -> HttpResponse:
    """Shuffles the queue. Only admin is permitted to do this."""
    if not user_manager.is_admin(request.user):
        return HttpResponseForbidden()
    playback.queue.shuffle()
    return HttpResponse()


@control
def remove_all(request: WSGIRequest) -> HttpResponse:
    """Empties the queue. Only admin is permitted to do this."""
    if not user_manager.is_admin(request.user):
        return HttpResponseForbidden()
    with transaction.atomic():
        playback.queue.all().delete()
    return HttpResponse()


@control
def prioritize(request: WSGIRequest) -> HttpResponse:
    """Prioritizes song by making it the first one in the queue."""
    key_param = request.POST.get("key")
    if key_param is None:
        return HttpResponseBadRequest()
    key = int(key_param)
    playback.queue.prioritize(key)
    return HttpResponse()


@control
def remove(request: WSGIRequest) -> HttpResponse:
    """Removes a song identified by the given key from the queue."""
    key_param = request.POST.get("key")
    if key_param is None:
        return HttpResponseBadRequest()
    key = int(key_param)
    try:
        removed = playback.queue.remove(key)
        # if we removed a song and it was added by autoplay,
        # we want it to be the new basis for autoplay
        if not removed.manually_requested:
            playback.handle_autoplay(removed.external_url or removed.title)
        else:
            playback.handle_autoplay()
    except models.QueuedSong.DoesNotExist:
        return HttpResponseBadRequest("song does not exist")
    return HttpResponse()


@control
def reorder(request: WSGIRequest) -> HttpResponse:
    """Reorders the queue.
    The song specified by element is inserted between prev and next."""
    prev_key_param = request.POST.get("prev")
    cur_key_param = request.POST.get("element")
    next_key_param = request.POST.get("next")
    if not cur_key_param:
        return HttpResponseBadRequest()
    if not prev_key_param:
        prev_key = None
    else:
        prev_key = int(prev_key_param)
    cur_key = int(cur_key_param)
    if not next_key_param:
        next_key = None
    else:
        next_key = int(next_key_param)
    try:
        playback.queue.reorder(prev_key, cur_key, next_key)
    except ValueError:
        return HttpResponseBadRequest("request on old state")
    return HttpResponse()


@csrf_exempt
@user_manager.tracked
def vote(request: WSGIRequest) -> HttpResponse:
    """Modify the vote-count of the given song by the given amount.
    If a song receives too many downvotes, it is removed."""
    key_param = request.POST.get("key")
    amount_param = request.POST.get("amount")
    if key_param is None or amount_param is None:
        return HttpResponseBadRequest()
    key = int(key_param)
    amount = int(amount_param)
    if amount < -2 or amount > 2 or amount == 0:
        return HttpResponseBadRequest()

    if storage.get("ip_checking") and not user_manager.try_vote(
        user_manager.get_client_ip(request), key, amount
    ):
        return HttpResponseBadRequest("nice try")

    if storage.get("color_indication") != storage.Privileges.nobody:
        user_manager.register_vote(request, key, amount)

    models.CurrentSong.objects.filter(queue_key=key).update(votes=F("votes") + amount)
    try:
        current_song = models.CurrentSong.objects.get()
        if (
            current_song.queue_key == key
            and current_song.votes
            <= -storage.get(  # pylint: disable=invalid-unary-operand-type
                "downvotes_to_kick"
            )
        ):
            _skip()
    except models.CurrentSong.DoesNotExist:
        pass

    removed = playback.queue.vote(
        key,
        amount,
        -storage.get("downvotes_to_kick"),  # pylint: disable=invalid-unary-operand-type
    )
    # if we removed a song by voting, and it was added by autoplay,
    # we want it to be the new basis for autoplay
    if removed is not None:
        if not removed.manually_requested:
            playback.handle_autoplay(removed.external_url or removed.title)
        else:
            playback.handle_autoplay()
    musiq.update_state()
    return HttpResponse()



from django.db import migrations, connection

if connection.vendor == "postgresql":
    from django.contrib.postgres.operations import TrigramExtension


class Migration(migrations.Migration):
    dependencies = [("core", "0014_auto_20211011_2041")]

    operations = [TrigramExtension()] if connection.vendor == "postgresql" else []



"""This module contains the celery app."""
import os
from threading import Thread
from typing import Any, Callable

from core.util import strtobool

app: Any

if strtobool(os.environ.get("DJANGO_NO_CELERY", "0")):
    # For debugging, celery's startup is quite slow, especially when reloading on every change.
    # Instead, use Threads to keep asynchronicity but with a much faster startup.
    CELERY_ACTIVE = False

    class MockCelery:
        """A mock class that starts threads instead of dispatching the tasks to celery workers."""

        def task(self, function: Callable) -> Callable:
            """This decorator mocks celery's delay function.
            This delay() creates a thread and starts it."""

            def delay(*args: Any, **kwargs: Any) -> None:
                thread = Thread(target=function, args=args, kwargs=kwargs, daemon=True)
                thread.start()

            # monkeypatch-add this method
            function.delay = delay  # type: ignore[attr-defined]

            return function

    def start() -> None:
        """MockCelery does not need to be initialized."""

    app = MockCelery()
else:
    CELERY_ACTIVE = True

    from celery import Celery

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "main.settings")

    app = Celery("core")

    app.config_from_object("django.conf:settings")

    class CeleryNotReachable(Exception):
        """Raised when celery should be reachable but is not."""

    def start() -> None:
        """Initializes celery."""
        # check if celery is up and wait for a maximum of 5 seconds
        for _ in range(10):
            if app.control.ping(timeout=0.5):
                break
        else:
            raise CeleryNotReachable("Celery worker pool not reachable. Is it running?")

        # stop running celery tasks from old django instance
        active_tasks = app.control.inspect().active()
        if active_tasks:
            for _, tasks in active_tasks.items():
                for task in tasks:
                    app.control.revoke(task_id=task["id"], terminate=True)



"""This module contains all programs that use the screen."""
import logging
import os
import subprocess
import time
from typing import List, TYPE_CHECKING, Optional

from django.conf import settings as conf
from core import redis
from core.lights import lights
from core.lights.exceptions import ScreenProgramStopped
from core.lights.programs import ScreenProgram

if TYPE_CHECKING:
    from core.lights.worker import DeviceManager


class Video(ScreenProgram):
    """Plays a given Video."""

    def __init__(self, manager: "DeviceManager", video: str, loop=False):
        super().__init__(manager, os.path.splitext(os.path.basename(video))[0])
        self.player: Optional[subprocess.Popen] = None
        if not os.path.isabs(video):
            video = os.path.join(conf.BASE_DIR, "resources", "videos", video)
        if not os.path.isfile(video):
            logging.warning("video %s does not exist", video)
            raise ValueError(f"video {video} does not exist")
        self.video = video
        self.loop = loop
        self.omxplayer = False

    def start(self) -> None:
        # omxplayer is preferred because it loops smoothly and is hardware accelerated
        try:
            args = ["omxplayer", self.video]
            if self.loop:
                args.append("--loop")
            self.player = subprocess.Popen(
                args, stdout=subprocess.DEVNULL, stdin=subprocess.PIPE
            )
            self.omxplayer = True
        except FileNotFoundError:
            # vlc is hardware accelerated, but shows a black screen between loops
            try:
                args = ["cvlc", "--fullscreen", "--no-video-title-show", self.video]
                if self.loop:
                    args.append("--loop")
                self.player = subprocess.Popen(args, stdout=subprocess.DEVNULL)
            except FileNotFoundError:
                # mplayer loops smoothly but is not hardware accelerated
                try:
                    args = ["mplayer", "-fs", self.video]
                    if self.loop:
                        args.append("-loop")
                        args.append("0")
                    self.player = subprocess.Popen(args, stdout=subprocess.DEVNULL)
                except FileNotFoundError as error:
                    raise ValueError("No video player found") from error

    def compute(self) -> None:
        if not self.player or self.player.poll() is not None:
            raise ScreenProgramStopped

    def stop(self) -> None:
        if not self.player:
            return
        if self.omxplayer:
            # for some reason omxplayer refuses to exit on sigterm from python
            self.player.communicate(b"q")
        else:
            self.player.terminate()


class Visualization(ScreenProgram):
    """Renders a visualization to the screen."""

    NUM_PARTICLES = 400
    FPS_MEASURE_WINDOW = 20.0  # seconds

    @staticmethod
    def get_variants() -> List[str]:
        """Returns all possible visualization variants by listing the available shaders."""
        # don't offer this feature on raspberry pi 3
        try:
            with open("/proc/device-tree/model", encoding="utf-8") as model_file:
                model = model_file.read()
                if model.startswith("Raspberry Pi 3"):
                    return []
        except FileNotFoundError:
            # we are not running on a raspberry pi
            pass

        try:
            import raveberry_visualization

            controller = raveberry_visualization.Controller()
            return controller.get_variants()
        except ModuleNotFoundError:
            return []

    def __init__(self, manager: "DeviceManager", variant: str) -> None:
        super().__init__(manager, variant)
        # __init__ is called for every variant reported by get_variants
        # that method acts as a guard that the module is installed
        import raveberry_visualization

        self.controller = raveberry_visualization.Controller()
        self.last_fps_check = time.time()

    def start(self) -> None:
        self.manager.utilities.cava.use()
        self.controller.start(
            self.name,
            self.manager.settings["ups"],
            Visualization.NUM_PARTICLES,
            Visualization.FPS_MEASURE_WINDOW,
        )

    def compute(self) -> None:
        now = time.time()
        if now - self.last_fps_check > Visualization.FPS_MEASURE_WINDOW / 2:
            self.last_fps_check = now
            current_fps = self.controller.get_fps()
            redis.put("current_fps", current_fps)
            if (
                self.manager.settings["dynamic_resolution"]
                and current_fps < 0.9 * self.manager.settings["ups"]
            ):
                self.manager.devices.screen.lower_resolution()
                # restart the program with the new resolution
                self.manager.restart_screen_program(sleep_time=2, has_lock=True)
            else:
                lights.update_state()
        if not self.controller.is_active():
            raise ScreenProgramStopped
        self.controller.set_parameters(
            self.manager.utilities.alarm.factor,
            self.manager.utilities.cava.current_frame,
        )

    def stop(self) -> None:
        self.controller.stop()
        self.manager.utilities.cava.release()



"""This module provides methods to access database settings."""
from ast import literal_eval
from typing import Union, Literal

from cachetools import TTLCache, cached

from core import models
from core.util import strtobool


class Interactivity:
    """An enum containing all possible interactivity level descriptions."""

    full_control = "Full Public Control"
    full_voting = "Up- and Downvoting"
    upvotes_only = "Upvotes Only"
    no_control = "No Control"


class Privileges:
    """An enum containing all privilege levels."""

    everybody = "Everybody"
    mod = "Mod and Admin"
    admin = "Admin Only"
    nobody = "Nobody"


PlatformEnabled = Literal
PlatformSuggestions = Literal
DeviceBrightness = Literal
DeviceMonochrome = Literal
DeviceProgram = Literal

# maps key to default and type of value
defaults = {
    # basic settings
    "interactivity": Interactivity.full_control,
    "ip_checking": False,
    "color_indication": Privileges.nobody,
    "color_offset": 0.0,
    "next_color_index": 0,
    "downvotes_to_kick": 2,
    "logging_enabled": True,
    "hashtags_active": True,
    "privileged_stream": False,
    "online_suggestions": True,
    "number_of_suggestions": 20,
    "connectivity_host": "1.1.1.1",
    "new_music_only": False,
    "enqueue_first": False,
    "song_cooldown": 0.0,
    "max_download_size": 0.0,
    "max_playlist_items": 10,
    "max_queue_length": 0,
    "additional_keywords": "",
    "forbidden_keywords": "",
    "people_to_party": 3,
    "alarm_probability": 0.0,
    "buzzer_cooldown": 1.0,
    "buzzer_success_probability": -1.0,
    # platforms
    "local_enabled": True,
    "youtube_enabled": True,
    "youtube_suggestions": 2,
    "spotify_enabled": False,
    "spotify_suggestions": 2,
    "spotify_username": "",
    "spotify_password": "",
    "spotify_device_client_id": "",
    "spotify_device_client_secret": "",
    "spotify_mopidy_client_id": "",
    "spotify_mopidy_client_secret": "",
    "spotify_redirect_uri": "",
    "spotify_authorized_url": "",
    "spotipy_token_info": "",
    "soundcloud_enabled": False,
    "soundcloud_suggestions": 2,
    "soundcloud_auth_token": "",
    "jamendo_enabled": False,
    "jamendo_suggestions": 2,
    "jamendo_client_id": "",
    # sound
    "feed_cava": True,
    "output": "",
    "backup_stream": "",
    # playback
    "paused": False,
    "volume": 1.0,
    "shuffle": False,
    "repeat": False,
    "autoplay": False,
    # lights
    "ups": 30.0,
    "fixed_color": (0.0, 0.0, 0.0),
    "program_speed": 0.5,
    "wled_led_count": 10,
    "wled_ip": "",
    "wled_port": 21324,
    # the concise, but not much shorter version:
    # **{
    #    k: v
    #    for k, v in list(
    #        chain.from_iterable(
    #            [
    #                (f"{device}_brightness", 1.0),
    #                (f"{device}_monochrome", False),
    #                (f"{device}_program", "Disabled"),
    #                (f"last_{device}_program", "Disabled"),
    #            ]
    #            for device in ["ring", "strip", "wled", "screen"]
    #        )
    #    )
    # },
    "ring_brightness": 1.0,
    "ring_monochrome": False,
    "ring_program": "Disabled",
    "last_ring_program": "Disabled",
    "strip_brightness": 1.0,
    "strip_monochrome": False,
    "strip_program": "Disabled",
    "last_strip_program": "Disabled",
    "wled_brightness": 1.0,
    "wled_monochrome": False,
    "wled_program": "Disabled",
    "last_wled_program": "Disabled",
    "screen_brightness": 1.0,
    "screen_monochrome": False,
    "screen_program": "Disabled",
    "last_screen_program": "Disabled",
    "initial_resolution": (0, 0),
    "dynamic_resolution": False,
}

# Settings change very rarely, cache them to reduce database roundtrips.
# This is especially advantageous for suggestions which check whether platforms are enabled.
# There is a data inconsistency issue when a setting is changed in one process.
# Only that process would flush its cache, others would retain the stale value.
# However, with the daphne setup there is currently only one process handling requests,
# and settings are never changed outside a request (especially not in a celery worker).
# So this is fine as long as no additional daphne (or other) workers are used.
# However, settings are accessed in both the lights and the playback worker.
# The lights flushes the cache in its update function.
# The playback worker accesses the "paused" state very often,
# so it is stored both in redis and the db.
# Alternatively, the cache flush could be communicated through redis.
cache: TTLCache = TTLCache(ttl=10, maxsize=128)


@cached(cache)
def get(key: str) -> Union[bool, int, float, str, tuple]:
    """This method returns the value for the given :param key:.
    Values of non-existing keys are set to their respective default value."""
    # values are stored as string in the database
    # cast the value to its respective type, defined by the default value, before returning it
    default = defaults[key]
    value = models.Setting.objects.get_or_create(
        key=key, defaults={"value": str(default)}
    )[0].value
    if type(default) is str:
        return str(value)
    if type(default) is int:
        return int(value)
    if type(default) is float:
        return float(value)
    if type(default) is bool:
        return strtobool(value)
    if type(default) is tuple:
        # evaluate the stored literal
        return literal_eval(value)
    raise ValueError(f"{key} not defined")


def put(key: str, value: Union[bool, int, float, str, tuple]) -> None:
    """This method sets the :param value: for the given :param key:."""
    default = defaults[key]
    setting = models.Setting.objects.get_or_create(
        key=key, defaults={"value": str(default)}
    )[0]
    setting.value = str(value)
    setting.save()
    cache.clear()



"""This module handles all settings related to wifi."""

from __future__ import annotations

import os
import subprocess

from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, HttpResponseBadRequest, JsonResponse

from core.settings.settings import control


@control
def available_ssids(_request: WSGIRequest) -> JsonResponse:
    """List all ssids that can currently be seen."""
    output = subprocess.check_output(
        ["sudo", "/usr/local/sbin/raveberry/list_available_ssids"]
    ).decode()
    ssids = output.split("\n")
    return JsonResponse(ssids[:-1], safe=False)


@control
def connect_to_wifi(request: WSGIRequest) -> HttpResponse:
    """Connect to a given ssid with the given password."""
    ssid = request.POST.get("ssid")
    password = request.POST.get("password")
    if ssid is None or password is None or ssid == "" or password == "":
        return HttpResponseBadRequest("Please provide both SSID and password")
    try:
        output = subprocess.check_output(
            ["sudo", "/usr/local/sbin/raveberry/connect_to_wifi", ssid, password]
        ).decode()
        return HttpResponse(output)
    except subprocess.CalledProcessError as error:
        output = error.output.decode()
        return HttpResponseBadRequest(output)


@control
def disable_homewifi(_request: WSGIRequest) -> None:
    """Disable homewifi function."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_homewifi"])


@control
def enable_homewifi(_request: WSGIRequest) -> None:
    """Enable homewifi function."""
    subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_homewifi"])


@control
def stored_ssids(_request: WSGIRequest) -> JsonResponse:
    """Return the list of ssids that this Raspberry Pi was connected to in the past."""
    output = subprocess.check_output(
        ["sudo", "/usr/local/sbin/raveberry/list_stored_ssids"]
    ).decode()
    ssids = output.split("\n")
    return JsonResponse(ssids[:-1], safe=False)


@control
def set_homewifi_ssid(request: WSGIRequest) -> HttpResponse:
    """Set the home network.
    The hotspot will not be created if connected to this wifi."""
    homewifi_ssid = request.POST.get("value")
    if homewifi_ssid is None:
        return HttpResponseBadRequest("homewifi ssid was not supplied.")
    with open(
        os.path.join(settings.BASE_DIR, "config/homewifi"), "w+", encoding="utf-8"
    ) as homewifi_file:
        homewifi_file.write(homewifi_ssid)
    return HttpResponse()



"""This module contains all url endpoints and maps them to their corresponding functions."""
import inspect
from typing import Any, List, Union

from django.core.handlers.wsgi import WSGIRequest
from django.urls import URLPattern, URLResolver, include, path
from django.views.generic import RedirectView

from core import api
from core import base
from core.lights import controller as lights_controller
from core.lights import lights
from core.musiq import controller as musiq_controller
from core.musiq import musiq
from core.musiq import suggestions
from core import network_info
from core.settings import analysis
from core.settings import basic
from core.settings import library
from core.settings import platforms
from core.settings import settings
from core.settings import sound
from core.settings import system
from core.settings import wifi
from core import state_handler

urlpatterns: List[Union[URLPattern, URLResolver]] = [
    path("", RedirectView.as_view(pattern_name="musiq", permanent=False), name="base"),
    path("musiq/", musiq.index, name="musiq"),
    path("lights/", lights.index, name="lights"),
    path("stream/", base.no_stream, name="no-stream"),
    path("network-info/", network_info.index, name="network-info"),
    path("settings/", settings.index, name="settings"),
    path("accounts/", include("django.contrib.auth.urls")),
    path("login/", RedirectView.as_view(pattern_name="login", permanent=False)),
    path("logged-in/", base.logged_in, name="logged-in"),
    path("logout/", RedirectView.as_view(pattern_name="logout", permanent=False)),
    path(
        "api/",
        include(
            [
                path("version/", api.version, name="version"),
                path(
                    "musiq/",
                    include(
                        [
                            path("post_song/", api.post_song, name="post_song"),
                            path("post-song/", api.post_song, name="post-song"),
                        ]
                    ),
                ),
            ]
        ),
    ),
]

urlmethods = [urlpattern.callback for urlpattern in urlpatterns]


def get_paths(objs: List[Any]) -> List[URLPattern]:
    """Iterates through the given objects and identifies all methods that serve http requests.
    A url pattern is generated for each of these methods,
    but only if no such path exists already in urlpatterns.
    Returns the list of url patterns."""
    paths = []
    for obj in objs:
        for name, method in inspect.getmembers(obj, inspect.isfunction):
            if name == "get_state":
                # the state url is an exception
                # it cannot be named after its method name, as every page has its own get_state
                continue
            if name.startswith("_"):
                # do not expose internal methods
                continue
            # use __annotations__ instead of get_type_hints, because the latter raised errors about
            # WSGIRequest not being defined in some configurations
            # string annotations are sufficient for our use case
            type_hints = method.__annotations__
            if "request" in type_hints:
                request_type = type_hints["request"]
            elif "_request" in type_hints:
                request_type = type_hints["_request"]
            else:
                continue
            if (
                request_type == "WSGIRequest" or issubclass(request_type, WSGIRequest)
            ) and method not in urlmethods:
                name = name.replace("_", "-")
                paths.append(path(name + "/", method, name=name))
    return paths


base_paths = get_paths([base])
musiq_paths = get_paths([musiq, musiq_controller, suggestions])
lights_paths = get_paths([lights_controller])
settings_paths = get_paths([basic, platforms, sound, wifi, library, analysis, system])

urlpatterns.append(
    path(
        "ajax/",
        include(
            [
                path("", include(base_paths)),
                path(
                    "musiq/state/",
                    state_handler.get_state,
                    {"module": musiq},
                    name="musiq-state",
                ),
                path("musiq/", include(musiq_paths)),
                path(
                    "lights/state/",
                    state_handler.get_state,
                    {"module": lights},
                    name="lights-state",
                ),
                path("lights/", include(lights_paths)),
                path(
                    "settings/state/",
                    state_handler.get_state,
                    {"module": settings},
                    name="settings-state",
                ),
                path("settings/", include(settings_paths)),
            ]
        ),
    )
)



"""This module contains the device superclass."""
from typing import cast

from core import redis
from core.redis import DeviceInitialized
from core.settings import storage
from core.settings.storage import DeviceBrightness, DeviceMonochrome
from core.lights.programs import LightProgram, Disabled


class Device:
    """A class representing a visualization device that Raveberry can control."""

    def __init__(self, manager, name) -> None:
        self.manager = manager
        self.name = name
        assert self.name in ["ring", "strip", "wled", "screen"]
        self.brightness = storage.get(cast(DeviceBrightness, f"{self.name}_brightness"))
        self.monochrome = storage.get(cast(DeviceMonochrome, f"{self.name}_monochrome"))
        self.initialized = False
        redis.put(cast(DeviceInitialized, f"{self.name}_initialized"), False)
        self.program: LightProgram = Disabled(manager)

    def load_program(self) -> None:
        """Load and activate this device's program from the database."""
        assert self.name in ["ring", "strip", "wled", "screen"]
        program_name = storage.get(cast(DeviceBrightness, f"{self.name}_program"))

        # only enable if the device is initialized
        if self.initialized:
            self.program = self.manager.programs[program_name]
        else:
            self.program = self.manager.utilities.disabled

        self.program.use()

    def clear(self) -> None:
        """Resets this device, clearing all visualization."""
        raise NotImplementedError()
